!---------------------------------------------------------------------------------------
! PREP_REGIONS: mod_strings
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> mod_strings
!!
!! Purpose:    Module for string functions.
!!
!---------------------------------------------------------------------------------------

module mod_strings

  use mod_var

  implicit none

  character(*), private, parameter :: lower_case = 'abcdefghijklmnopqrstuvwxyz' 
  character(*), private, parameter :: upper_case = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ' 

  public :: to_upper, to_lower, str_replace, split

  contains

  ! --------------------------------------------------
  ! to uppercase
  ! --------------------------------------------------
  !> to_upper
  !! 
  !! Purpose:  Converts a string to all uppercase
  !!           characters.
  !!
  ! --------------------------------------------------

    function to_upper( string_in ) result ( string_out ) 
 
      implicit none

      character(*), intent(in)      :: string_in 
      character( len( string_in ) ) :: string_out 
      integer                       :: i, n 

      string_out = string_in 
      do i = 1, len( string_out ) 
        ! find location of letter in lower case constant string 
        n = index( lower_case, string_in(i:i) ) 
        ! if current substring is a lower case letter, make it upper case 
        if ( n.ne.0 ) string_out(i:i) = upper_case(n:n) 
      end do

    end function to_upper 

  ! --------------------------------------------------
  ! to lowercase
  ! --------------------------------------------------
  !> to_lower
  !!
  !! Purpose:  Converts a string all lowercase
  !!           characters.
  !!
  ! --------------------------------------------------

    function to_lower( string_in ) result ( string_out )         

      implicit none

      character(*), intent(in)      :: string_in      
      character( len( string_in ) ) :: string_out      
      integer                       :: i, n

      string_out = string_in        
      do i = 1, len( string_out )   
        ! find location of letter in upper case constant string 
        n = index( upper_case, string_in(i:i) )      
        ! if current substring is a upper case letter, make it upper case 
        if ( n.ne.0 ) string_out(i:i) = lower_case(n:n)        
      end do

    end function to_lower

  ! --------------------------------------------------
  ! replace substring in string
  ! --------------------------------------------------
  !> str_replace
  !!
  !! Purpose:  Replaces a regular string pattern in
  !!           a string.
  !!
  ! --------------------------------------------------

    function str_replace( string_in, substr, repstr ) result ( string_out )

      implicit none

      character(*), intent(in)      :: string_in
      character(*), intent(in)      :: substr
      character(*), intent(in)      :: repstr
      character(max_path_len)       :: string_out
      character(max_path_len)       :: strtmp1, strtmp2
      integer                       :: strlen, substrlen, replen, n

      strlen = len_trim(string_in)
      replen = len_trim(repstr)
      substrlen = len_trim(substr)
      ! find first position of substr in string_in 
      n = index( string_in, substr, back=.false. )
      if ((n-1).gt.0) then
        strtmp1 = string_in(1:n-1)
      else
        ! empty string
        strtmp1 = ''
        n = 1
      endif
!      print*, 'str_replace first pos n, strtmp1 = ',n, strtmp1
      ! find last position of substr in string_in 
      n = index( string_in, substr, back=.true. )
      strtmp2 = string_in(n+substrlen:strlen)
!      print*, 'str_replace last pos n, strtmp2 = ',n, strtmp2
      string_out = trim(strtmp1)//trim(repstr)//trim(strtmp2)

    end function str_replace

    ! --------------------------------------------------
    ! split string
    ! --------------------------------------------------

    subroutine split(str,delims,before,sep)

    ! Routine finds the first instance of a character from 'delims' in the
    ! the string 'str'. The characters before the found delimiter are
    ! output in 'before'. The characters after the found delimiter are
    ! output in 'str'. The optional output character 'sep' contains the 
    ! found delimiter. A delimiter in 'str' is treated like an ordinary 
    ! character if it is preceded by a backslash (\). If the backslash 
    ! character is desired in 'str', then precede it with another backslash.

    character(len=*) :: str, delims, before
    character,optional :: sep
    logical :: pres
    character :: ch,cha
    integer :: i, k, ibsl, ipos, iposa, lenstr

    pres=present(sep)
    str=adjustl(str)
    lenstr=len_trim(str)
    if(lenstr == 0) return        ! string str is empty
    k=0
    ibsl=0                        ! backslash initially inactive
    before=' '
    do i=1,lenstr
      ch=str(i:i)
      if(ibsl == 1) then          ! backslash active
        k=k+1
        before(k:k)=ch
        ibsl=0
        cycle
      end if
      if(ch == '\') then          ! backslash with backslash inactive
        k=k+1
        before(k:k)=ch
        ibsl=1
        cycle
      end if
      ipos=index(delims,ch)
      if(ipos == 0) then          ! character is not a delimiter
        k=k+1
        before(k:k)=ch
        cycle
      end if
      if(ch /= ' ') then          ! character is a delimiter that is not a space
        str=str(i+1:)
        if(pres) sep=ch
        exit
      end if
      cha=str(i+1:i+1)            ! character is a space delimiter
      iposa=index(delims,cha)
      if(iposa > 0) then          ! next character is a delimiter
        str=str(i+2:)
        if(pres) sep=cha
        exit
      else
        str=str(i+1:)
        if(pres) sep=ch
        exit
      end if
    end do
    if(i >= lenstr) str=''
    str=trim(str)
    before=trim(before)
    return

    end subroutine split

  ! --------------------------------------------------

end module mod_strings


