!---------------------------------------------------------------------------------------
! PREP_SATELLITE: mod_save
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2022, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> mod_save
!!
!! Purpose:    Writes retrieval and releases data to netcdf files
!!
!---------------------------------------------------------------------------------------

module mod_save

  use mod_settings
  use mod_var
  use netcdf

  implicit none
  private

  public :: check, write_releases, write_retrieval

  contains

  ! --------------------------------------------------
  ! check
  ! --------------------------------------------------

  subroutine check(status)

    integer, intent(in) :: status

    if ( status.ne.nf90_noerr ) then
      write(logid,*) trim(nf90_strerror(status))
      stop
    end if

  end subroutine check

  ! --------------------------------------------------
  ! write_releases
  !
  ! total column SRR calculated as 
  !
  ! SRR_col = sum(a_i * SRR_i * vdryair_i)
  !
  ! where
  !   a_i        ith layer of column averaging kernel
  !   vdryair_i  dry air weighting in ith layer
  !   SRR_col    column SRR in units of s.mol/kg 
  !              (after dividing by height of surface 
  !              layer) 
  !
  ! --------------------------------------------------

  subroutine write_releases(settings,numbox,nlayer,ncoord,xpoints,ypoints,zpoint1,zpoint2,&
                             idate,itime,cak_out,vdryair_out,specnum,units)

    type (settings_t),               intent(in) :: settings
    integer,                         intent(in) :: numbox, nlayer, ncoord
    real, dimension(maxretr,ncoord), intent(in) :: xpoints, ypoints
    real, dimension(maxretr,nlayer), intent(in) :: zpoint1, zpoint2
    integer, dimension(maxretr),     intent(in) :: idate, itime
    real, dimension(maxretr,nlayer), intent(in) :: cak_out, vdryair_out
    integer,                         intent(in) :: specnum
    character(len=6), dimension(4),  intent(in) :: units

    character(len=8)                            :: adate
    character(max_name_len)                     :: filename, pathout
    integer                                     :: ncid, nrid, nlid, ncoid
    integer                                     :: xptid, yptid, zpt1id, zpt2id
    integer                                     :: dateid, timeid, cakid, dairid, npartid

    write(adate,fmt='(I8.8)') idate(1)
    filename = 'releases_'//trim(settings%satname)//'_'//adate//'.nc'
    pathout = trim(settings%path_options)//adate//'/options/'
    call system('mkdir '//trim(pathout))
    write(logid,*) 'writing file: ',trim(pathout)//trim(filename)

    ! create file 
    call check( nf90_create(trim(pathout)//trim(filename), nf90_clobber, ncid) )

    ! define dimensions and variables
    call check( nf90_def_dim(ncid, 'retrieval', numbox, nrid) )
    call check( nf90_def_dim(ncid, 'nlayer', nlayer, nlid) )
    call check( nf90_def_dim(ncid, 'ncoord', ncoord, ncoid) )
    call check( nf90_def_var(ncid, 'xpoints', nf90_real, (/nrid, ncoid/), xptid) )
    call check( nf90_put_att(ncid, xptid, 'units', 'degrees_east') )
    call check( nf90_def_var(ncid, 'ypoints', nf90_real, (/nrid, ncoid/), yptid) )
    call check( nf90_put_att(ncid, yptid, 'units', 'degrees_north') )
    call check( nf90_def_var(ncid, 'zpoint1', nf90_real, (/nrid, nlid/), zpt1id) )
    call check( nf90_put_att(ncid, zpt1id, 'units', 'Pa') )
    call check( nf90_def_var(ncid, 'zpoint2', nf90_real, (/nrid, nlid/), zpt2id) )
    call check( nf90_put_att(ncid, zpt2id, 'units', 'Pa') )
    call check( nf90_def_var(ncid, 'idate', nf90_int, nrid, dateid) )
    call check( nf90_put_att(ncid, dateid, 'units', 'none') )
    call check( nf90_def_var(ncid, 'itime', nf90_int, nrid, timeid) )
    call check( nf90_put_att(ncid, timeid, 'units', 'none') )
    call check( nf90_def_var(ncid, 'cak', nf90_real, (/nrid, nlid/), cakid) )
    call check( nf90_put_att(ncid, cakid, 'units', 'none') )
    call check( nf90_def_var(ncid, 'dryair', nf90_real, (/nrid, nlid/), dairid) )
    call check( nf90_put_att(ncid, dairid, 'units', trim(units(1))) )
    call check( nf90_put_att(ncid, nf90_global, 'specnum', specnum) )
    call check( nf90_put_att(ncid, nf90_global, 'mass', settings%mass) )
    call check( nf90_put_att(ncid, nf90_global, 'npart', settings%npart) )

    ! exit define mode
    call check( nf90_enddef(ncid) )

    ! insert variables
    call check( nf90_put_var(ncid, xptid, xpoints(1:numbox,:)) )
    call check( nf90_put_var(ncid, yptid, ypoints(1:numbox,:)) )
    call check( nf90_put_var(ncid, zpt1id, zpoint1(1:numbox,:)) )
    call check( nf90_put_var(ncid, zpt2id, zpoint2(1:numbox,:)) )
    call check( nf90_put_var(ncid, dateid, idate(1:numbox)) )
    call check( nf90_put_var(ncid, timeid, itime(1:numbox)) )
    call check( nf90_put_var(ncid, cakid, cak_out(1:numbox,:)) )
    call check( nf90_put_var(ncid, dairid, vdryair_out(1:numbox,:)) )
  
    ! close file
    call check( nf90_close(ncid) )

  end subroutine write_releases

  ! --------------------------------------------------
  ! write_retrieval
  !
  ! model total column is:
  !
  ! vmod = cpri - cakpri + SRR_col * flux 
  !
  ! and total column mixing ratio (ppbv) is
  !
  ! xmod = vmod/vdryair
  !
  ! where 
  !   cpri     sum(vpri_i)
  !   cakpri   sum(a_i*vpri_i)
  !   vdryair  total column dry air
  !
  ! --------------------------------------------------

  subroutine write_retrieval(settings,numbox,ncoord,xpoints,ypoints,idate,itime,&
                             cpri,cakpri,vmr_out,verr_out,cdryair,albedo_out,units)

    type (settings_t),               intent(in) :: settings 
    integer,                         intent(in) :: numbox, ncoord
    real, dimension(maxretr,ncoord), intent(in) :: xpoints, ypoints
    integer, dimension(maxretr),     intent(in) :: idate, itime
    real, dimension(maxretr),        intent(in) :: vmr_out, verr_out, albedo_out
    real, dimension(maxretr),        intent(in) :: cpri, cakpri, cdryair
    character(len=6), dimension(4),  intent(in) :: units

    character(max_name_len)                     :: filename
    character(len=8)                            :: adate
    integer                                     :: ncid, nrid, nlid, ncoid
    integer                                     :: xptid, yptid, dateid, timeid
    integer                                     :: cprid, cakprid, vmrid, verrid, albid, cdairid    

    write(adate,fmt='(I8.8)') idate(1)
    filename = 'retrieval_'//trim(settings%satname)//'_'//adate//'.nc'
    write(logid,*) 'writing file: ',trim(settings%path_obsout)//trim(filename)

    ! create file
    call check( nf90_create(trim(settings%path_obsout)//trim(filename), nf90_clobber, ncid) )

    ! define dimensions and variables
    call check( nf90_def_dim(ncid, 'retrieval', numbox, nrid) )
    call check( nf90_def_dim(ncid, 'ncoord', ncoord, ncoid) )
    call check( nf90_def_var(ncid, 'xpoints', nf90_real, (/nrid, ncoid/), xptid) )
    call check( nf90_put_att(ncid, xptid, 'units', 'degrees_east') )
    call check( nf90_def_var(ncid, 'ypoints', nf90_real, (/nrid, ncoid/), yptid) )
    call check( nf90_put_att(ncid, yptid, 'units', 'degrees_north') )
    call check( nf90_def_var(ncid, 'idate', nf90_int, nrid, dateid) )
    call check( nf90_put_att(ncid, dateid, 'units', 'none') )
    call check( nf90_def_var(ncid, 'itime', nf90_int, nrid, timeid) )
    call check( nf90_put_att(ncid, timeid, 'units', 'none') )
    call check( nf90_def_var(ncid, 'cpri', nf90_real, nrid, cprid) )
    call check( nf90_put_att(ncid, cprid, 'units', trim(units(3))) )
    call check( nf90_def_var(ncid, 'cakpri', nf90_real, nrid, cakprid) )
    call check( nf90_put_att(ncid, cakprid, 'units', trim(units(3))) )
    call check( nf90_def_var(ncid, 'vmr', nf90_real, nrid, vmrid) )
    call check( nf90_put_att(ncid, vmrid, 'units', trim(units(4))) )
    call check( nf90_def_var(ncid, 'vmr_err', nf90_real, nrid, verrid) )
    call check( nf90_put_att(ncid, verrid, 'units', trim(units(4))) )
    call check( nf90_def_var(ncid, 'albedo_swir', nf90_real, nrid, albid) )
    call check( nf90_put_att(ncid, albid, 'units', 'none') )
    call check( nf90_def_var(ncid, 'cdryair', nf90_real, nrid, cdairid) )
    call check( nf90_put_att(ncid, cdairid, 'units', trim(units(2))) )

    ! end define mode
    call check( nf90_enddef(ncid) )

    ! insert variables
    call check( nf90_put_var(ncid, xptid, xpoints(1:numbox,:)) )
    call check( nf90_put_var(ncid, yptid, ypoints(1:numbox,:)) )
    call check( nf90_put_var(ncid, dateid, idate(1:numbox)) )
    call check( nf90_put_var(ncid, timeid, itime(1:numbox)) )
    call check( nf90_put_var(ncid, cprid, cpri(1:numbox)) )
    call check( nf90_put_var(ncid, cakprid, cakpri(1:numbox)) )
    call check( nf90_put_var(ncid, vmrid, vmr_out(1:numbox)) )
    call check( nf90_put_var(ncid, verrid, verr_out(1:numbox)) )
    call check( nf90_put_var(ncid, albid, albedo_out(1:numbox)) )
    call check( nf90_put_var(ncid, cdairid, cdryair(1:numbox)) )
  
    ! close file
    call check( nf90_close(ncid) )

  end subroutine write_retrieval

  ! --------------------------------------------------

end module mod_save

