!---------------------------------------------------------------------------------------
! FLEXINVERT: mod_covar
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> mod_covar
!!
!! Purpose:    Module for constructing the error covariance data structure
!!
!---------------------------------------------------------------------------------------


module mod_covar

  use mod_var

  implicit none
  private

  public :: covar_t, alloc_covar, dealloc_covar, matrix_oper

  type :: covar_t
 
    real(kind=8), dimension(:),   allocatable :: evals       ! eigen values of error covariance matrix
    real(kind=8), dimension(:,:), allocatable :: evecs       ! eigen vectors of error covariance matrix
    real(kind=8), dimension(:,:), allocatable :: cort        ! temporal correlation matrix
    real(kind=8), dimension(:,:), allocatable :: icort       ! inverse of temporal correlation matrix
    real(kind=8), dimension(:,:), allocatable :: sqcort      ! square root of temporal correlation matrix
    real(kind=8), dimension(:,:), allocatable :: isqcort     ! inverse of sqrt of temporal correlation matrix
    real(kind=8), dimension(:,:), allocatable :: corcini     ! spatial correlation matrix for scalars of initial mixing ratio
    real(kind=8), dimension(:,:), allocatable :: cortcini    ! temporal correlation matrix for scalars of initial mixing ratio
    real(kind=8), dimension(:,:), allocatable :: icortcini   ! inverse of temporal correlation matrix of scalars
    real(kind=8), dimension(:,:), allocatable :: sqcortcini  ! square root of temporal correlation matrix of scalars
    real(kind=8), dimension(:,:), allocatable :: isqcortcini ! inverse of sqrt of temporal correlation matrix of scalars
    real(kind=8), dimension(:,:), allocatable :: icorcini    ! inverse of covariance matrix of scalars
    real(kind=8), dimension(:,:), allocatable :: sqcorcini   ! square root of covariance matrix of scalars
    real(kind=8), dimension(:,:), allocatable :: isqcorcini  ! inverse of sqrt of covariance matrix of scalars

  end type

  contains

  ! --------------------------------------------------
  ! alloc_covar
  ! --------------------------------------------------

  subroutine alloc_covar(covar)

    use mod_var 

    type (covar_t), intent(in out)       :: covar
    integer                              :: ierr

    allocate( covar%evals(neig), stat = ierr )
    if ( ierr.ne.0 ) then
      write(logid,*) 'ERROR alloc_covar: not enough memory'
      stop
    endif
    allocate( covar%evecs(ndvar,neig), stat = ierr )
    if ( ierr.ne.0 ) then
      write(logid,*) 'ERROR alloc_covar: not enough memory'
      stop
    endif
    allocate( covar%cort(ntstate,ntstate), stat = ierr )
    if ( ierr.ne.0 ) then
      write(logid,*) 'ERROR alloc_covar: not enough memory'
      stop
    endif
    allocate( covar%icort(ntstate,ntstate), stat = ierr )
    if ( ierr.ne.0 ) then
      write(logid,*) 'ERROR alloc_covar: not enough memory'
      stop
    endif
    allocate( covar%sqcort(ntstate,ntstate), stat = ierr )
    if ( ierr.ne.0 ) then
      write(logid,*) 'ERROR alloc_covar: not enough memory'
      stop
    endif
    allocate( covar%isqcort(ntstate,ntstate), stat = ierr )
    if ( ierr.ne.0 ) then
      write(logid,*) 'ERROR alloc_covar: not enough memory'
      stop
    endif
    allocate( covar%corcini(ncini,ncini), stat = ierr )
    if ( ierr.ne.0 ) then
      write(logid,*) 'ERROR alloc_covar: not enough memory'
      stop
    endif
    allocate( covar%sqcorcini(ncini,ncini), stat = ierr )
    if ( ierr.ne.0 ) then
      write(logid,*) 'ERROR alloc_covar: not enough memory'
      stop
    endif
    allocate( covar%icorcini(ncini,ncini), stat = ierr )
    if ( ierr.ne.0 ) then
      write(logid,*) 'ERROR alloc_covar: not enough memory'
      stop
    endif
    allocate( covar%isqcorcini(ncini,ncini), stat = ierr )
    if ( ierr.ne.0 ) then
      write(logid,*) 'ERROR alloc_covar: not enough memory'
      stop
    endif
    allocate( covar%cortcini(ntcini,ntcini), stat = ierr )
    if ( ierr.ne.0 ) then
      write(logid,*) 'ERROR alloc_covar: not enough memory'
      stop
    endif
    allocate( covar%sqcortcini(ntcini,ntcini), stat = ierr )
    if ( ierr.ne.0 ) then
      write(logid,*) 'ERROR alloc_covar: not enough memory'
      stop
    endif
    allocate( covar%icortcini(ntcini,ntcini), stat = ierr )
    if ( ierr.ne.0 ) then
      write(logid,*) 'ERROR alloc_covar: not enough memory'
      stop
    endif
    allocate( covar%isqcortcini(ntcini,ntcini), stat = ierr )
    if ( ierr.ne.0 ) then
      write(logid,*) 'ERROR alloc_covar: not enough memory'
      stop
    endif

  end subroutine alloc_covar

  ! --------------------------------------------------
  ! dealloc_covar
  ! --------------------------------------------------

  subroutine dealloc_covar(covar)

    type (covar_t), intent(in out)       :: covar

    if ( allocated(covar%evals) )      deallocate(covar%evals)
    if ( allocated(covar%evecs) )      deallocate(covar%evecs)
    if ( allocated(covar%cort) )       deallocate(covar%cort)
    if ( allocated(covar%icort) )      deallocate(covar%icort)
    if ( allocated(covar%sqcort) )     deallocate(covar%sqcort)
    if ( allocated(covar%isqcort) )    deallocate(covar%isqcort)
    if ( allocated(covar%corcini) )    deallocate(covar%corcini)
    if ( allocated(covar%sqcorcini) )  deallocate(covar%sqcorcini)
    if ( allocated(covar%icorcini) )   deallocate(covar%icorcini)
    if ( allocated(covar%isqcorcini) ) deallocate(covar%isqcorcini)
    if ( allocated(covar%cortcini) )   deallocate(covar%cortcini)
    if ( allocated(covar%sqcortcini) ) deallocate(covar%sqcortcini)
    if ( allocated(covar%icortcini) )  deallocate(covar%icortcini)
    if ( allocated(covar%isqcortcini)) deallocate(covar%isqcortcini)

  end subroutine dealloc_covar

  ! --------------------------------------------------
  ! matrix_oper
  ! --------------------------------------------------
  !
  !> matrix_oper
  !!
  !! Purpose:  calculate eigendecomposition of matrix
  !!           and uses this to calculate matrix
  !!           square root and inverse
  !!
  ! --------------------------------------------------

  subroutine matrix_oper(ndim, mat, imat, sqmat, isqmat)

    integer, intent(in)                            :: ndim
    real(kind=8), dimension(ndim,ndim), intent(in) :: mat
    real(kind=8), dimension(ndim,ndim), intent(out):: imat, sqmat, isqmat

    integer                                        :: lwork, liwork
    integer, dimension(:), allocatable             :: isuppz
    integer, dimension(:), allocatable             :: iwork
    real(kind=8), dimension(:), allocatable        :: work
    real(kind=8), dimension(:,:), allocatable      :: zwork
    real(kind=8), dimension(:), allocatable        :: evals
    real(kind=8), dimension(:,:), allocatable      :: evecs
    real(kind=8), dimension(ndim)                  :: tmp
    integer                                        :: ierr, info, neval, ind
    integer                                        :: i, j

    ! mat = U*D*U^(T)
    lwork = 26*ndim
    liwork = 10*ndim

    allocate( work(lwork), stat=ierr )
    if ( ierr.ne.0 ) stop 'ERROR mod_covar: not enough memory'
    allocate( iwork(liwork), stat=ierr )
    if ( ierr.ne.0 ) stop 'ERROR mod_covar: not enough memory'
    allocate( isuppz(2*ndim), stat=ierr )
    if ( ierr.ne.0 ) stop 'ERROR mod_covar: not enough memory'
    allocate( evecs(ndim,ndim), stat=ierr )
    if ( ierr.ne.0 ) stop 'ERROR mod_covar: not enough memory'
    allocate( evals(ndim), stat=ierr )
    if ( ierr.ne.0 ) stop 'ERROR mod_covar: not enough memory'
    allocate( zwork(ndim,ndim), stat=ierr )
    if ( ierr.ne.0 ) stop 'ERROR mod_covar: not enough memory'

    zwork = mat
    call dsyevr('V','A','L',ndim,zwork,ndim,1.,1.,1,1,0.,neval,evals,evecs,ndim,isuppz,work,lwork,iwork,liwork,info)
    if ( info.ne.0 ) then
      write(logid,*) 'ERROR error_cov: eigenvalue decomposition'
      stop
    endif

    ! inverse of matrix
    imat(:,:) = 0d0
    do i = 1, ndim
      do j = 1, ndim
        tmp(j) = evecs(i,j)*1d0/evals(j)
      end do
      imat(i,:) = matmul(tmp,transpose(evecs))
    end do

    ! square root of matrix
    sqmat(:,:) = 0d0
    do i = 1, ndim
      do j = 1, ndim
        tmp(j) = evecs(i,j)*sqrt(evals(j))
      end do
      sqmat(i,:) = matmul(tmp,transpose(evecs))
    end do

    ! inverse of square root of matrix
    isqmat(:,:) = 0d0
    do i = 1, ndim
      do j = 1, ndim
        tmp(j) = evecs(i,j)*1d0/sqrt(evals(j))
      end do
      isqmat(i,:) = matmul(tmp,transpose(evecs))
    end do

    deallocate(work)
    deallocate(zwork)
    deallocate(iwork)
    deallocate(isuppz)
    deallocate(evals)
    deallocate(evecs)

  end subroutine matrix_oper

  ! --------------------------------------------------

end module mod_covar




